# __version__.py
# autogenerated by keats 0.2.25
__version__ = "0.1.5a6"
__name__ = "pydent"
__title__ = "pydent"
__authors__ = [
    "jvrana <justin.vrana@gmail.com>",
    "Ben Keller <bjkeller@uw.edu>",
    "Eric Klavins <klavins@uw.edu>",
]
__repo__ = None
__homepage__ = "https://www.github.com/klavinslab/trident"
__description__ = "Aquarium's Python API for planning, executing, and analyzing scientific experiments."
