from abc import ABC


class SessionABC(ABC):

    Account = None
    AllowableFieldType = None
    Budget = None
    Code = None
    Collection = None
    DataAssociation = None
    FieldType = None
    FieldValue = None
    Group = None
    Invoice = None
    Item = None
    Job = None
    JobAssociation = None
    Library = None
    Membership = None
    ObjectType = None
    Operation = None
    OperationType = None
    PartAssociation = None
    Plan = None
    PlanAssociation = None
    Sample = None
    SampleType = None
    Upload = None
    User = None
    UserBudgetAssociation = None
    Wire = None
